# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional

import attr

from mozanalysis.experiment import AnalysisBasis


@attr.s(frozen=True, slots=True)
class DataSource:
    """Represents a table or view, from which Metrics may be defined.

    Args:
        name (str): Name for the Data Source. Used in sanity metric
            column names.
        from_expr (str): FROM expression - often just a fully-qualified
            table name. Sometimes a subquery. May contain the string
            ``{dataset}`` which will be replaced with an app-specific
            dataset for Glean apps. If the expression is templated
            on dataset, default_dataset is mandatory.
        experiments_column_type (str or None): Info about the schema
            of the table or view:

            * 'simple': There is an ``experiments`` column, which is an
              (experiment_slug:str -> branch_name:str) map.
            * 'native': There is an ``experiments`` column, which is an
              (experiment_slug:str -> struct) map, where the struct
              contains a ``branch`` field, which is the branch as a
              string.
            * None: There is no ``experiments`` column, so skip the
              sanity checks that rely on it. We'll also be unable to
              filter out pre-enrollment data from day 0 in the
              experiment.
        client_id_column (str, optional): Name of the column that
            contains the ``client_id`` (join key). Defaults to
            'client_id'.
        submission_date_column (str, optional): Name of the column
            that contains the submission date (as a date, not
            timestamp). Defaults to 'submission_date'.
        default_dataset (str, optional): The value to use for
            `{dataset}` in from_expr if a value is not provided
            at runtime. Mandatory if from_expr contains a
            `{dataset}` parameter.
    """

    name = attr.ib(validator=attr.validators.instance_of(str))
    _from_expr = attr.ib(validator=attr.validators.instance_of(str))
    experiments_column_type = attr.ib(default="simple", type=str)
    client_id_column = attr.ib(default="client_id", type=str)
    submission_date_column = attr.ib(default="submission_date", type=str)
    default_dataset = attr.ib(default=None, type=Optional[str])

    EXPERIMENT_COLUMN_TYPES = (None, "simple", "native", "glean")

    @experiments_column_type.validator
    def _check_experiments_column_type(self, attribute, value):
        if value not in self.EXPERIMENT_COLUMN_TYPES:
            raise ValueError(
                f"experiments_column_type {repr(value)} must be one of: "
                f"{repr(self.EXPERIMENT_COLUMN_TYPES)}"
            )

    @default_dataset.validator
    def _check_default_dataset_provided_if_needed(self, attribute, value):
        self.from_expr_for(None)

    def from_expr_for(self, dataset: Optional[str]) -> str:
        """Expands the ``from_expr`` template for the given dataset.
        If ``from_expr`` is not a template, returns ``from_expr``.

        Args:
            dataset (str or None): Dataset name to substitute
                into the from expression.
        """
        effective_dataset = dataset or self.default_dataset
        if effective_dataset is None:
            try:
                return self._from_expr.format()
            except Exception as e:
                raise ValueError(
                    f"{self.name}: from_expr contains a dataset template but no value was provided."  # noqa:E501
                ) from e
        return self._from_expr.format(dataset=effective_dataset)

    @property
    def experiments_column_expr(self):
        if self.experiments_column_type is None:
            return ""

        elif self.experiments_column_type == "simple":
            return """AND (
                    ds.{submission_date} != e.enrollment_date
                    OR `mozfun.map.get_key`(
                        ds.experiments, '{experiment_slug}'
                    ) IS NOT NULL
                )"""

        elif self.experiments_column_type == "native":
            return """AND (
                    ds.{submission_date} != e.enrollment_date
                    OR `mozfun.map.get_key`(
                        ds.experiments, '{experiment_slug}'
                    ).branch IS NOT NULL
            )"""

        elif self.experiments_column_type == "glean":
            return """AND (
                    ds.{submission_date} != e.enrollment_date
                    OR `mozfun.map.get_key`(
                        ds.ping_info.experiments, '{experiment_slug}'
                    ).branch IS NOT NULL
                )"""

        else:
            raise ValueError

    def build_query(
        self,
        metric_list,
        time_limits,
        experiment_slug,
        from_expr_dataset=None,
        analysis_basis=AnalysisBasis.ENROLLMENTS,
        exposure_signal=None,
    ):
        """Return a nearly-self contained SQL query.

        This query does not define ``enrollments`` but otherwise could
        be executed to query all metrics from this data source.
        """
        return """
        SELECT
            e.client_id,
            e.branch,
            e.analysis_window_start,
            e.analysis_window_end,
            e.num_exposure_events,
            e.exposure_date,
            {metrics}
        FROM enrollments e
            LEFT JOIN {from_expr} ds
                ON ds.{client_id} = e.client_id
                AND ds.{submission_date} BETWEEN '{fddr}' AND '{lddr}'
                AND ds.{submission_date} BETWEEN
                    DATE_ADD(e.{date}, interval e.analysis_window_start day)
                    AND DATE_ADD(e.{date}, interval e.analysis_window_end day)
                {ignore_pre_enroll_first_day}
        GROUP BY
            e.client_id,
            e.branch,
            e.num_exposure_events,
            e.exposure_date,
            e.analysis_window_start,
            e.analysis_window_end""".format(
            client_id=self.client_id_column,
            submission_date=self.submission_date_column,
            from_expr=self.from_expr_for(from_expr_dataset),
            fddr=time_limits.first_date_data_required,
            lddr=time_limits.last_date_data_required,
            metrics=",\n            ".join(
                "{se} AS {n}".format(
                    se=m.select_expr.format(experiment_slug=experiment_slug), n=m.name
                )
                for m in metric_list
            ),
            date="exposure_date"
            if analysis_basis == AnalysisBasis.EXPOSURES and exposure_signal is None
            else "enrollment_date",
            ignore_pre_enroll_first_day=self.experiments_column_expr.format(
                submission_date=self.submission_date_column,
                experiment_slug=experiment_slug,
            ),
        )

    def get_sanity_metrics(self, experiment_slug):
        if self.experiments_column_type is None:
            return []

        elif self.experiments_column_type == "simple":
            return [
                Metric(
                    name=self.name + "_has_contradictory_branch",
                    data_source=self,
                    select_expr=agg_any(
                        """`mozfun.map.get_key`(
                ds.experiments, '{experiment_slug}'
            ) != e.branch"""
                    ),
                ),
                Metric(
                    name=self.name + "_has_non_enrolled_data",
                    data_source=self,
                    select_expr=agg_any(
                        """`mozfun.map.get_key`(
                ds.experiments, '{experiment_slug}'
            ) IS NULL""".format(
                            experiment_slug=experiment_slug
                        )
                    ),
                ),
            ]

        elif self.experiments_column_type == "native":
            return [
                Metric(
                    name=self.name + "_has_contradictory_branch",
                    data_source=self,
                    select_expr=agg_any(
                        """`mozfun.map.get_key`(
                ds.experiments, '{experiment_slug}'
            ).branch != e.branch"""
                    ),
                ),
                Metric(
                    name=self.name + "_has_non_enrolled_data",
                    data_source=self,
                    select_expr=agg_any(
                        """`mozfun.map.get_key`(
                ds.experiments, '{experiment_slug}'
            ).branch IS NULL""".format(
                            experiment_slug=experiment_slug
                        )
                    ),
                ),
            ]

        elif self.experiments_column_type == "glean":
            return [
                Metric(
                    name=self.name + "_has_contradictory_branch",
                    data_source=self,
                    select_expr=agg_any(
                        """`mozfun.map.get_key`(
                ds.ping_info.experiments, '{experiment_slug}'
            ).branch != e.branch"""
                    ),
                ),
                Metric(
                    name=self.name + "_has_non_enrolled_data",
                    data_source=self,
                    select_expr=agg_any(
                        """`mozfun.map.get_key`(
                ds.ping_info.experiments, '{experiment_slug}'
            ).branch IS NULL""".format(
                            experiment_slug=experiment_slug
                        )
                    ),
                ),
            ]

        else:
            raise ValueError


@attr.s(frozen=True, slots=True)
class Metric:
    """Represents an experiment metric.

    Needs to be combined with an analysis window to be measurable!

    Args:
        name (str): A slug; uniquely identifies this metric in tables
        data_source (DataSource): where to find the metric
        select_expr (str): a SQL snippet representing a clause of a SELECT
            expression describing how to compute the metric; must include an
            aggregation function since it will be GROUPed BY client_id
            and branch
        friendly_name (str): A human-readable dashboard title for this metric
        description (str): A paragraph of Markdown-formatted text describing
            what the metric measures, to be shown on dashboards
    """

    name = attr.ib(type=str)
    data_source = attr.ib(type=DataSource)
    select_expr = attr.ib(type=str)
    friendly_name = attr.ib(type=Optional[str], default=None)
    description = attr.ib(type=Optional[str], default=None)
    bigger_is_better = attr.ib(type=bool, default=True)


def agg_sum(select_expr):
    """Return a SQL fragment for the sum over the data, with 0-filled nulls."""
    return "COALESCE(SUM({}), 0)".format(select_expr)


def agg_any(select_expr):
    """Return the logical OR, with FALSE-filled nulls."""
    return "COALESCE(LOGICAL_OR({}), FALSE)".format(select_expr)


def agg_histogram_mean(select_expr):
    """Produces an expression for the mean of an unparsed histogram."""
    return f"""SAFE_DIVIDE(
                SUM(CAST(JSON_EXTRACT_SCALAR({select_expr}, "$.sum") AS int64)),
                SUM((SELECT SUM(value) FROM UNNEST(mozfun.hist.extract({select_expr}).values)))
            )"""  # noqa



import pytest
from cheap_lint import sql_lint

import mozanalysis.metrics.desktop as mad
import mozanalysis.metrics.fenix
import mozanalysis.metrics.firefox_ios
import mozanalysis.segments.desktop as msd
from mozanalysis.experiment import AnalysisBasis, AnalysisWindow, Experiment, TimeLimits
from mozanalysis.exposure import ExposureSignal
from mozanalysis.metrics import Metric
from mozanalysis.segments import Segment, SegmentDataSource


def test_time_limits_validates():
    # Mainly check that the validation is running at all
    # No need to specify the same checks twice(?)
    with pytest.raises(TypeError):
        TimeLimits()

    with pytest.raises(AssertionError):
        TimeLimits(
            first_enrollment_date="2019-01-05",
            last_enrollment_date="2019-01-05",
            analysis_windows=(AnalysisWindow(1, 1),),
            first_date_data_required="2019-01-01",  # Before enrollments
            last_date_data_required="2019-01-01",
        )


def test_time_limits_create1():
    # When we have complete data for 2019-01-14...
    # ...We have 14 dates of data for those who enrolled on the 1st
    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        analysis_start_days=0,
        analysis_length_dates=14,
    )

    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-01"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 13
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-01-14"


def test_time_limits_create2():
    # We don't have 14 dates of data for an 8-day cohort:
    with pytest.raises(ValueError):
        TimeLimits.for_single_analysis_window(
            first_enrollment_date="2019-01-01",
            last_date_full_data="2019-01-14",
            analysis_start_days=0,
            analysis_length_dates=14,
            num_dates_enrollment=8,
        )

    # We don't have 15 full dates of data for any users
    with pytest.raises(AssertionError):
        TimeLimits.for_single_analysis_window(
            first_enrollment_date="2019-01-01",
            last_date_full_data="2019-01-14",
            analysis_start_days=0,
            analysis_length_dates=15,
        )


def test_time_limits_create3():
    # For the 8-day cohort We have enough data for a 7 day window
    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        analysis_start_days=0,
        analysis_length_dates=7,
        num_dates_enrollment=8,
    )
    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-08"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 6
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-01-14"


def test_time_limits_create4():
    # Or a 2 day window
    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        analysis_start_days=0,
        analysis_length_dates=2,
        num_dates_enrollment=8,
    )
    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-08"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 1
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-01-09"


def test_time_limits_create5():
    # But not an 8 day window
    with pytest.raises(ValueError):
        TimeLimits.for_single_analysis_window(
            first_enrollment_date="2019-01-01",
            last_date_full_data="2019-01-14",
            analysis_start_days=0,
            analysis_length_dates=8,
            num_dates_enrollment=8,
        )


def test_time_limits_create6():
    # Of course the flexi-experiment has data for a 1 day window
    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        analysis_start_days=0,
        analysis_length_dates=1,
    )
    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-14"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 0
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-01-14"


def test_time_limits_create7():
    # If the analysis starts later, so does the data source
    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        analysis_start_days=7,
        analysis_length_dates=1,
    )
    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-07"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 7
    assert tl.analysis_windows[0].end == 7
    assert tl.first_date_data_required == "2019-01-08"
    assert tl.last_date_data_required == "2019-01-14"


def test_ts_time_limits_create1():
    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        time_series_period="daily",
        num_dates_enrollment=8,
    )

    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-08"
    assert len(tl.analysis_windows) == 7
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 0
    assert tl.analysis_windows[6].start == 6
    assert tl.analysis_windows[6].end == 6
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-01-14"


def test_ts_time_limits_create2():
    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-08"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 6
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-01-14"


def test_ts_time_limits_create3():
    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-15",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-08"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 6
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-01-14"


def test_ts_time_limits_create4():
    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-02-15",
        time_series_period="28_day",
        num_dates_enrollment=8,
    )

    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-08"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 27
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-02-04"


def test_ts_time_limits_create_not_enough_data():
    with pytest.raises(ValueError):
        TimeLimits.for_ts(
            first_enrollment_date="2019-01-01",
            last_date_full_data="2019-01-13",
            time_series_period="weekly",
            num_dates_enrollment=8,
        )


def test_time_limits_has_right_date_in_error_message():
    msg_re = r"until we have data for 2020-03-30."
    with pytest.raises(ValueError, match=msg_re):
        TimeLimits.for_single_analysis_window(
            first_enrollment_date="2020-03-03",
            last_date_full_data="2020-03-23",
            analysis_start_days=0,
            analysis_length_dates=21,
            num_dates_enrollment=8,
        )


def test_analysis_window_validates_start():
    AnalysisWindow(0, 1)
    with pytest.raises(AssertionError):
        AnalysisWindow(-1, 1)


def test_analysis_window_validates_end():
    AnalysisWindow(5, 5)
    with pytest.raises(AssertionError):
        AnalysisWindow(5, 4)


def test_query_not_detectably_malformed():
    exp = Experiment("slug", "2019-01-01", 8)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl, enrollments_query_type="normandy"
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=[],
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)


def test_megaquery_not_detectably_malformed():
    exp = Experiment("slug", "2019-01-01", 8)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl, enrollments_query_type="normandy"
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=[m for m in mad.__dict__.values() if isinstance(m, mad.Metric)],
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)


def test_segments_megaquery_not_detectably_malformed():
    exp = Experiment("slug", "2019-01-01", 8)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl,
        segment_list=[s for s in msd.__dict__.values() if isinstance(s, msd.Segment)],
        enrollments_query_type="normandy",
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=[m for m in mad.__dict__.values() if isinstance(m, mad.Metric)],
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)


def test_app_id_propagates():
    exp = Experiment("slug", "2019-01-01", 8, app_id="my_cool_app")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    sds = SegmentDataSource(
        name="cool_data_source",
        from_expr="`moz-fx-data-shared-prod`.{dataset}.cool_table",
        default_dataset="org_mozilla_firefox",
    )

    segment = Segment(
        name="cool_segment",
        select_expr="COUNT(*)",
        data_source=sds,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl, segment_list=[segment], enrollments_query_type="fenix-fallback"
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=[
            m
            for m in mozanalysis.metrics.fenix.__dict__.values()
            if isinstance(m, Metric)
        ],
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)

    assert "org_mozilla_firefox" not in enrollments_sql
    assert "my_cool_app" in enrollments_sql

    sql_lint(metrics_sql)


def test_query_not_detectably_malformed_fenix_fallback():
    exp = Experiment("slug", "2019-01-01", 8)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl, enrollments_query_type="fenix-fallback"
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=[],
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)


def test_firefox_ios_app_id_propagation():
    exp = Experiment("slug", "2019-01-01", 8, app_id="my_cool_app")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    sds = SegmentDataSource(
        name="cool_data_source",
        from_expr="`moz-fx-data-shared-prod`.{dataset}.cool_table",
        default_dataset="org_mozilla_ios_firefox",
    )

    segment = Segment(
        name="cool_segment",
        select_expr="COUNT(*)",
        data_source=sds,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl,
        segment_list=[segment],
        enrollments_query_type="glean-event",
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=[
            m
            for m in mozanalysis.metrics.firefox_ios.__dict__.values()
            if isinstance(m, Metric)
        ],
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)

    assert "org_mozilla_ios_firefox" not in enrollments_sql
    assert "my_cool_app" in enrollments_sql

    sql_lint(metrics_sql)


def test_firefox_klar_app_id_propagation():
    exp = Experiment("slug", "2019-01-01", 8, app_id="my_cool_app")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    sds = SegmentDataSource(
        name="cool_data_source",
        from_expr="`moz-fx-data-shared-prod`.{dataset}.cool_table",
        default_dataset="org_mozilla_klar",
    )

    segment = Segment(
        name="cool_segment",
        select_expr="COUNT(*)",
        data_source=sds,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl,
        segment_list=[segment],
        enrollments_query_type="glean-event",
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=[
            m
            for m in mozanalysis.metrics.klar_android.__dict__.values()
            if isinstance(m, Metric)
        ],
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)

    assert "org_mozilla_klar" not in enrollments_sql
    assert "my_cool_app" in enrollments_sql

    sql_lint(metrics_sql)


def test_firefox_ios_klar_app_id_propagation():
    exp = Experiment("slug", "2019-01-01", 8, app_id="my_cool_app")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    sds = SegmentDataSource(
        name="cool_data_source",
        from_expr="`moz-fx-data-shared-prod`.{dataset}.cool_table",
        default_dataset="org_mozilla_ios_klar",
    )

    segment = Segment(
        name="cool_segment",
        select_expr="COUNT(*)",
        data_source=sds,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl,
        segment_list=[segment],
        enrollments_query_type="glean-event",
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=[
            m
            for m in mozanalysis.metrics.klar_ios.__dict__.values()
            if isinstance(m, Metric)
        ],
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)

    assert "org_mozilla_ios_klar" not in enrollments_sql
    assert "my_cool_app" in enrollments_sql

    sql_lint(metrics_sql)


def test_exposure_query():
    exp = Experiment("slug", "2019-01-01", 8, app_id="my_cool_app")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollment_sql = exp.build_enrollments_query(
        time_limits=tl,
        enrollments_query_type="glean-event",
    )

    sql_lint(enrollment_sql)

    assert "exposures" in enrollment_sql


def test_exposure_signal_query():
    exp = Experiment("slug", "2019-01-01", 8, app_id="my_cool_app")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollment_sql = exp.build_enrollments_query(
        time_limits=tl,
        enrollments_query_type="glean-event",
        exposure_signal=ExposureSignal(
            name="exposures",
            data_source=mozanalysis.metrics.fenix.baseline,
            select_expr="metrics.counter.events_total_uri_count > 0",
            friendly_name="URI visited exposure",
            description="Exposed when URI visited",
        ),
    )

    sql_lint(enrollment_sql)

    assert "exposures" in enrollment_sql
    assert "metrics.counter.events_total_uri_count > 0" in enrollment_sql


def test_exposure_signal_query_custom_windows():
    exp = Experiment("slug", "2019-01-01", 8, app_id="my_cool_app")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollment_sql = exp.build_enrollments_query(
        time_limits=tl,
        enrollments_query_type="glean-event",
        exposure_signal=ExposureSignal(
            name="exposures",
            data_source=mozanalysis.metrics.fenix.baseline,
            select_expr="metrics.counter.events_total_uri_count > 0",
            friendly_name="URI visited exposure",
            description="Exposed when URI visited",
            window_start=1,
            window_end=3,
        ),
    )

    sql_lint(enrollment_sql)

    assert "exposures" in enrollment_sql
    assert "metrics.counter.events_total_uri_count > 0" in enrollment_sql
    assert "DATE_ADD('2019-01-01', INTERVAL 1 DAY)" in enrollment_sql
    assert "DATE_ADD('2019-01-01', INTERVAL 3 DAY)" in enrollment_sql


def test_metrics_query_based_on_exposure():
    exp = Experiment("slug", "2019-01-01", 8)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl, enrollments_query_type="fenix-fallback"
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=[
            m
            for m in mozanalysis.metrics.fenix.__dict__.values()
            if isinstance(m, Metric)
        ],
        time_limits=tl,
        enrollments_table="enrollments",
        analysis_basis=AnalysisBasis.EXPOSURES,
    )

    sql_lint(metrics_sql)

    assert "e.exposure_date" in metrics_sql


def test_metrics_query_with_exposure_signal_custom_windows():
    exp = Experiment("slug", "2019-01-01", 8)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl, enrollments_query_type="fenix-fallback"
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=[
            m
            for m in mozanalysis.metrics.fenix.__dict__.values()
            if isinstance(m, Metric)
        ],
        time_limits=tl,
        enrollments_table="enrollments",
        analysis_basis=AnalysisBasis.EXPOSURES,
        exposure_signal=ExposureSignal(
            name="exposures",
            data_source=mozanalysis.metrics.fenix.baseline,
            select_expr="metrics.counter.events_total_uri_count > 0",
            friendly_name="URI visited exposure",
            description="Exposed when URI visited",
            window_start=1,
            window_end=3,
        ),
    )

    sql_lint(metrics_sql)

    assert "DATE_ADD('2019-01-01', INTERVAL 1 DAY)" in metrics_sql
    assert "DATE_ADD('2019-01-01', INTERVAL 3 DAY)" in metrics_sql


def test_metrics_query_with_exposure_signal():
    exp = Experiment("slug", "2019-01-01", 8)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl, enrollments_query_type="fenix-fallback"
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=[
            m
            for m in mozanalysis.metrics.fenix.__dict__.values()
            if isinstance(m, Metric)
        ],
        time_limits=tl,
        enrollments_table="enrollments",
        analysis_basis=AnalysisBasis.EXPOSURES,
        exposure_signal=ExposureSignal(
            name="exposures",
            data_source=mozanalysis.metrics.fenix.baseline,
            select_expr="metrics.counter.events_total_uri_count > 0",
            friendly_name="URI visited exposure",
            description="Exposed when URI visited",
        ),
    )

    sql_lint(metrics_sql)

    assert "DATE_ADD('2019-01-01', INTERVAL 0 DAY)" in metrics_sql
    assert "DATE_ADD('2019-01-08', INTERVAL 0 DAY)" in metrics_sql



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import Optional

import attr

from mozanalysis.metrics import DataSource


@attr.s(frozen=True, slots=True)
class ExposureSignal:
    """Represents a signal that determines an exposure has happened.

    The optional ``window_start`` and ``window_end`` parameters define
    the window of data used to determine whether each client has been exposed.
    ``window_start`` and ``window_end`` are integers, representing the number
    of days before or after the first enrollment date.
    If ``window_start`` is set to None, then by default the first enrollment date
    is used. If ``window_end`` is set to None, then by default the last enrollment
    date is used.


    Args:
        name (str): A slug; uniquely identifies this exposure signal in tables
        data_source (DataSource): data source to query from
        select_expr (str): a SQL snippet representing a clause of a SELECT
            expression describing how to compute the exposure signal; must
            be a boolean condition
        friendly_name (str): A human-readable dashboard title for this exposure signal
        description (str): A paragraph of Markdown-formatted text describing
            what the exposure signal represents, to be shown on dashboards
        window_start (int): Optional, see above
        window_end (int): Optional, see above
    """

    name = attr.ib(type=str)
    data_source = attr.ib(type=DataSource)
    select_expr = attr.ib(type=str)
    friendly_name = attr.ib(type=Optional[str], default=None)
    description = attr.ib(type=Optional[str], default=None)
    window_start = attr.ib(type=Optional[int], default=None)
    window_end = attr.ib(type=Optional[int], default=None)

    def build_query(
        self,
        time_limits,
    ):
        """Return a nearly self-contained query for determining exposures.

        This query does not define ``enrollments`` but otherwise could
        be executed to query all exposures based on the exposure metric
        from this data source.
        """
        return """SELECT
            e.client_id,
            e.branch,
            MIN(ds.submission_date) AS exposure_date,
            COUNT(ds.submission_date) AS num_exposure_events
        FROM raw_enrollments e
            LEFT JOIN (
                SELECT
                    {client_id} AS client_id,
                    {submission_date} AS submission_date
                FROM {from_expr}
                WHERE {submission_date}
                    BETWEEN DATE_ADD('{date_start}', INTERVAL {window_start} DAY)
                    AND DATE_ADD('{date_end}', INTERVAL {window_end} DAY)
                    AND {exposure_signal}
            ) AS ds
            ON ds.client_id = e.client_id AND
                ds.submission_date >= e.enrollment_date
        GROUP BY
            e.client_id,
            e.branch""".format(
            client_id=self.data_source.client_id_column,
            submission_date=self.data_source.submission_date_column,
            from_expr=self.data_source.from_expr_for(None),
            date_start=time_limits.first_enrollment_date,
            date_end=time_limits.first_enrollment_date
            if self.window_end
            else time_limits.last_enrollment_date,
            window_start=self.window_start or 0,
            window_end=self.window_end or 0,
            exposure_signal=self.select_expr,
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import re

from google.api_core.exceptions import Conflict
from google.cloud import bigquery


def sanitize_table_name_for_bq(table_name):
    of_good_character_but_possibly_verbose = re.sub(r"[^a-zA-Z_0-9]", "_", table_name)

    if len(of_good_character_but_possibly_verbose) <= 1024:
        return of_good_character_but_possibly_verbose

    return (
        of_good_character_but_possibly_verbose[:500]
        + "___"
        + of_good_character_but_possibly_verbose[-500:]
    )


class BigQueryContext:
    """Holds a BigQuery client, and some configuration.

    Args:
        dataset_id (str): Your `BigQuery dataset id`_.
        project_id (str, optional): Your BigQuery project_id.
            Defaults to the DS team's project.

    .. _BigQuery dataset id: https://cloud.google.com/bigquery/docs/datasets
    """

    def __init__(self, dataset_id, project_id="moz-fx-data-bq-data-science"):
        self.dataset_id = dataset_id
        self.project_id = project_id
        self.client = bigquery.Client(project=project_id)

    def run_query(self, sql, results_table=None):
        """Run a query and return the result.

        If ``results_table`` is provided, then save the results
        into there (or just query from there if it already exists).
        Returns a ``google.cloud.bigquery.table.RowIterator``

        Args:
            sql (str): A SQL query.
            results_table (str, optional): A table name, not including
                a project_id or dataset_id. The table name is used as a
                cache key (if the table already exists, we ignore ``sql``
                and return the table's contents), so it is wise for
                ``results_table`` to include a hash of ``sql``.
        """
        if not results_table:
            return self.client.query(sql).result()

        try:
            full_res = self.client.query(
                sql,
                job_config=bigquery.QueryJobConfig(
                    destination=self.client.dataset(self.dataset_id).table(
                        results_table
                    )
                ),
            ).result()
            print("Saved into", results_table)
            return full_res

        except Conflict:
            print("Table already exists. Reusing", results_table)
            return self.client.list_rows(self.fully_qualify_table_name(results_table))

    def fully_qualify_table_name(self, table_name):
        """Given a table name, return it fully qualified."""
        return "{project_id}.{dataset_id}.{full_table_name}".format(
            project_id=self.project_id,
            dataset_id=self.dataset_id,
            full_table_name=table_name,
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from textwrap import dedent

from mozanalysis.metrics import DataSource, Metric, agg_any, agg_sum

#: DataSource: The clients_daily table.
clients_daily = DataSource(
    name="clients_daily",
    from_expr="mozdata.telemetry.clients_daily",
)

#: DataSource: The `search_clients_engines_sources_daily`_ table.
#: This table unpacks search counts from the main ping;
#: it contains one row per (client_id, submission_date, engine, source).
#:
#: .. _`search_clients_engines_sources_daily`: https://docs.telemetry.mozilla.org/
#:    datasets/search/search_clients_engines_sources_daily/reference.html
search_clients_engines_sources_daily = DataSource(
    name="search_clients_engines_sources_daily",
    from_expr="mozdata.search.search_clients_engines_sources_daily",
    experiments_column_type=None,
)

#: DataSource: A clone of `search_clients_engines_sources_daily`.
#: Exists for backwards compatibility; new uses should use the new name.
search_clients_daily = search_clients_engines_sources_daily

#: DataSource: The main_summary table.
main_summary = DataSource(
    name="main_summary", from_expr="mozdata.telemetry.main_summary"
)

#: DataSource: The events table.
events = DataSource(
    name="events",
    from_expr="mozdata.telemetry.events",
    experiments_column_type="native",
)

#: DataSource: Normandy events; a subset of the events table.
#: The telemetry.events table is clustered by event_category.
#: Normandy accounts for about 10% of event volume, so this dramatically
#: reduces bytes queried compared to counting rows from the generic events DataSource.
normandy_events = DataSource(
    name="normandy_events",
    from_expr="""(
        SELECT
            *
        FROM mozdata.telemetry.events
        WHERE event_category = 'normandy'
    )""",
    experiments_column_type="native",
)

#: DataSource: The telemetry_stable.main_v4 ping table.
#: The main_v4 table is what backs the telemetry.main view.
#: Referencing the table directly helps us stay under the BigQuery
#: query complexity budget.
main = DataSource(
    name="main",
    from_expr="""(
                SELECT
                    *,
                    DATE(submission_timestamp) AS submission_date,
                    environment.experiments
                FROM `moz-fx-data-shared-prod`.telemetry_stable.main_v4
            )""",
    experiments_column_type="native",
)

#: DataSource: The telemetry.crash ping table.
crash = DataSource(
    name="crash",
    from_expr="""(
                SELECT
                    *,
                    DATE(submission_timestamp) AS submission_date,
                    environment.experiments
                FROM mozdata.telemetry.crash
            )""",
    experiments_column_type="native",
)

#: DataSource: The ``messaging_system.cfr`` table.
cfr = DataSource(
    name="cfr",
    from_expr="""(
                SELECT
                    *,
                    DATE(submission_timestamp) AS submission_date
                FROM `moz-fx-data-derived-datasets`.messaging_system.cfr
            )""",
    experiments_column_type="native",
)

#: DataSource: The ``activity_stream.events`` table.
activity_stream_events = DataSource(
    name="activity_stream_events",
    from_expr="""(
                SELECT
                    *,
                    DATE(submission_timestamp) AS submission_date
                FROM mozdata.activity_stream.events
            )""",
    experiments_column_type="native",
)


#: Metric: ...
active_hours = Metric(
    name="active_hours",
    data_source=clients_daily,
    select_expr=agg_sum("active_hours_sum"),
    friendly_name="Active hours",
    description=dedent(
        """\
        Measures the amount of time (in 5-second increments) during which
        Firefox received user input from a keyboard or mouse. The Firefox
        window does not need to be focused.
    """
    ),
)

#: Metric: ...
uri_count = Metric(
    name="uri_count",
    data_source=clients_daily,
    select_expr=agg_sum("scalar_parent_browser_engagement_total_uri_count_sum"),
    friendly_name="URIs visited",
    description=dedent(
        """\
        Counts the total number of URIs visited.
        Includes within-page navigation events (e.g. to anchors).
    """
    ),
)

#: Metric: ...
search_count = Metric(
    name="search_count",
    data_source=search_clients_engines_sources_daily,
    select_expr=agg_sum("sap"),
    friendly_name="SAP searches",
    description=dedent(
        """\
        Counts the number of searches a user performed through Firefox's
        Search Access Points.
        Learn more in the
        [search data documentation](https://docs.telemetry.mozilla.org/datasets/search.html).
    """  # noqa:E501
    ),
)

#: Metric: ...
tagged_search_count = Metric(
    name="tagged_search_count",
    data_source=search_clients_engines_sources_daily,
    select_expr=agg_sum("tagged_sap"),
    friendly_name="Tagged SAP searches",
    description=dedent(
        """\
        Counts the number of searches a user performed through Firefox's
        Search Access Points that were submitted with a partner code
        and were potentially revenue-generating.
        Learn more in the
        [search data documentation](https://docs.telemetry.mozilla.org/datasets/search.html).
    """  # noqa:E501
    ),
)

#: Metric: ...
tagged_follow_on_search_count = Metric(
    name="tagged_follow_on_search_count",
    data_source=search_clients_engines_sources_daily,
    select_expr=agg_sum("tagged_follow_on"),
    friendly_name="Tagged follow-on searches",
    description=dedent(
        """\
        Counts the number of follow-on searches with a Mozilla partner tag.
        These are additional searches that users performed from a search engine
        results page after executing a tagged search through a SAP.
        Learn more in the
        [search data documentation](https://docs.telemetry.mozilla.org/datasets/search.html).
    """  # noqa:E501
    ),
)

#: Metric: ...
ad_clicks = Metric(
    name="ad_clicks",
    data_source=search_clients_engines_sources_daily,
    select_expr=agg_sum("ad_click"),
    friendly_name="Ad clicks",
    description=dedent(
        """\
        Counts clicks on ads on search engine result pages with a Mozilla
        partner tag.
    """
    ),
)

#: Metric: ...
searches_with_ads = Metric(
    name="searches_with_ads",
    data_source=search_clients_engines_sources_daily,
    select_expr=agg_sum("search_with_ads"),
    friendly_name="Search result pages with ads",
    description=dedent(
        """\
        Counts search result pages served with advertising.
        Users may not actually see these ads thanks to e.g. ad-blockers.
        Learn more in the
        [search analysis documentation](https://mozilla-private.report/search-analysis-docs/book/in_content_searches.html).
    """  # noqa:E501
    ),
)

#: Metric: ...
organic_search_count = Metric(
    name="organic_search_count",
    data_source=search_clients_engines_sources_daily,
    select_expr=agg_sum("organic"),
    friendly_name="Organic searches",
    description=dedent(
        """\
        Counts organic searches, which are searches that are _not_ performed
        through a Firefox SAP and which are not monetizable.
        Learn more in the
        [search data documentation](https://docs.telemetry.mozilla.org/datasets/search.html).
    """  # noqa:E501
    ),
)

#: Metric: ...
unenroll = Metric(
    name="unenroll",
    data_source=normandy_events,
    select_expr=agg_any(
        """
                event_category = 'normandy'
                AND event_method = 'unenroll'
                AND event_string_value = '{experiment_slug}'
            """
    ),
    friendly_name="Unenrollments",
    description=dedent(
        """\
        Counts the number of clients with an experiment unenrollment event.
    """
    ),
    bigger_is_better=False,
)

#: Metric: ...
view_about_logins = Metric(
    name="view_about_logins",
    data_source=events,
    select_expr=agg_any(
        """
                event_method = 'open_management'
                AND event_category = 'pwmgr'
            """
    ),
    friendly_name="about:logins viewers",
    description=dedent(
        """\
        Counts the number of clients that viewed about:logins.
    """
    ),
)

#: Metric: ...
view_about_protections = Metric(
    name="view_about_protections",
    data_source=events,
    select_expr=agg_any(
        """
                event_method = 'show'
                AND event_object = 'protection_report'
            """
    ),
    friendly_name="about:protections viewers",
    description=dedent(
        """\
        Counts the number of clients that viewed about:protections.
    """
    ),
)

#: Metric: ...
connect_fxa = Metric(
    name="connect_fxa",
    data_source=events,
    select_expr=agg_any(
        """
                event_method = 'connect'
                AND event_object = 'account'
            """
    ),
    friendly_name="Connected FxA",
    description=dedent(
        """\
        Counts the number of clients that took action to connect to FxA.
        This does not include clients that were already connected to FxA at
        the start of the experiment and remained connected.
    """
    ),
)

#: Metric: Pocket organic rec clicks in New Tab
pocket_rec_clicks = Metric(
    name="pocket_rec_clicks",
    data_source=activity_stream_events,
    select_expr="""COUNTIF(
                event = 'CLICK'
                AND source = 'CARDGRID'
                AND JSON_EXTRACT_SCALAR(value, '$.card_type') = 'organic'
            )""",
    friendly_name="Clicked Pocket organic recs in New Tab",
    description=dedent(
        """\
         Counts the number of Pocket rec clicks made by each client.
    """
    ),
)

#: Metric: Pocket sponsored content clicks in New Tab
pocket_spoc_clicks = Metric(
    name="pocket_spoc_clicks",
    data_source=activity_stream_events,
    select_expr="""COUNTIF(
                event = 'CLICK'
                AND source = 'CARDGRID'
                AND JSON_EXTRACT_SCALAR(value, '$.card_type') = 'spoc'
            )""",
    friendly_name="Clicked Pocket sponsored content in New Tab",
    description=dedent(
        """\
         Counts the number of Pocket sponsored content clicks made by each client.
    """
    ),
)

#: Metric: ...
days_of_use = Metric(
    name="days_of_use",
    data_source=clients_daily,
    select_expr="COUNT(ds.submission_date)",
    friendly_name="Days of use",
    description="The number of days in the interval that each client sent a main ping.",
)

#: Metric: Clicks to disable Pocket in New Tab
disable_pocket_clicks = Metric(
    name="disable_pocket_clicks",
    data_source=activity_stream_events,
    select_expr="""COUNTIF(
                event = 'PREF_CHANGED'
                AND source = 'TOP_STORIES'
                AND JSON_EXTRACT_SCALAR(value, '$.status') = 'false'
            )""",
    friendly_name="Disabled Pocket in New Tab",
    description=dedent(
        """\
         Counts the number of clicks to disable Pocket in New Tab made by each client.
    """
    ),
)

#: Metric: Clicks to disable Pocket sponsored content in New Tab
disable_pocket_spocs_clicks = Metric(
    name="disable_pocket_spocs_clicks",
    data_source=activity_stream_events,
    select_expr="""COUNTIF(
                event = 'PREF_CHANGED'
                AND source = 'POCKET_SPOCS'
                AND JSON_EXTRACT_SCALAR(value, '$.status') = 'false'
            )""",
    friendly_name="Disabled Pocket sponsored content in New Tab",
    description=dedent(
        """\
         Counts the number of clicks to disable Pocket sponsored content
         in New Tab made by each client.
    """
    ),
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import pandas as pd
import pytest

import mozanalysis.bayesian_stats as mabs
import mozanalysis.bayesian_stats.binary as mabsb


def test_aggregate_col():
    df = pd.DataFrame(
        [
            ["ctl", 1, 0],
            ["ctl", 1, 0],
            ["ctl", 0, 1],
            ["tst", 1, 0],
            ["tst", 0, 1],
        ],
        columns=["branch", "val1", "val2"],
    )

    res1 = mabsb.aggregate_col(df, "val1")
    assert res1.loc["ctl", "num_enrollments"] == 3
    assert res1.loc["tst", "num_enrollments"] == 2
    assert res1.loc["ctl", "num_conversions"] == 2
    assert res1.loc["tst", "num_conversions"] == 1

    res2 = mabsb.aggregate_col(df, "val2")
    assert res2.loc["ctl", "num_enrollments"] == 3
    assert res2.loc["tst", "num_enrollments"] == 2
    assert res2.loc["ctl", "num_conversions"] == 1
    assert res2.loc["tst", "num_conversions"] == 1


def test_summarize_one_branch_vs_samples():
    s = pd.Series([30, 80], index=["num_conversions", "num_enrollments"])

    ppf = mabsb.summarize_one_branch_from_agg(s, quantiles=[0.5, 0.41])

    samples = mabsb.get_samples(
        s.to_frame().T, "num_enrollments", "num_conversions", 100000
    )

    res = mabs.summarize_one_branch_samples(samples.iloc[:, 0], quantiles=[0.5, 0.41])

    assert ppf["0.5"] == pytest.approx(res["0.5"], abs=0.001)
    assert ppf["0.41"] == pytest.approx(res["0.41"], abs=0.001)
    assert ppf["mean"] == pytest.approx(res["mean"], abs=0.001)


def test_compare_branches():
    # The stats has been checked above; this just exercises the remaining
    # business logic

    df = pd.DataFrame(
        [
            ["ctl", 1, 0],
            ["ctl", 1, 0],
            ["ctl", 0, 1],
            ["tst", 1, 0],
            ["tst", 0, 1],
            ["3rd", 1, 0],
            ["3rd", 0, 1],
        ],
        columns=["branch", "val1", "val2"],
    )

    res = mabsb.compare_branches(df, "val2", ref_branch_label="ctl")

    assert set(res["individual"].keys()) == set(df.branch.unique())
    assert set(res["comparative"].keys()) == set(df.branch.unique()) - set(["ctl"])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np
import pandas as pd
import pytest

import mozanalysis.bayesian_stats.bayesian_bootstrap as mabsbb
import mozanalysis.bayesian_stats.binary as mabsbin
import mozanalysis.frequentist_stats.bootstrap as mafsb


def test_bootstrap_vs_beta(spark_context_or_none):
    num_enrollments = 10000
    fake_data = pd.Series(np.zeros(num_enrollments))
    fake_data[:300] = 1

    boot_res = mafsb.bootstrap_one_branch(fake_data, sc=spark_context_or_none)
    beta_res = mabsbin.summarize_one_branch_from_agg(
        pd.Series(
            {
                # `-1` to simulate Beta(0, 0) improper prior, closer to
                # bootstrap for quantiles (i think?)
                "num_enrollments": len(fake_data) - 1,
                "num_conversions": fake_data.sum() - 1,
            }
        )
    )

    for l in boot_res.index:
        # Bootstrapped quantiles are discretized based on the number of enrollments,
        # which sets `abs`.
        #
        # Is `num_samples` large enough to consistently achieve results that
        # match the beta model to within the accuracy of this discrete limit?
        # Not quite. So we backed it off a bit and ask for the bootstrapped result
        # to be within 1.9 quanta of the beta result - which was enough for a
        # percentile bootstrap
        assert boot_res.loc[l] == pytest.approx(
            beta_res.loc[l],
            # abs=1.9 is usually good enough with a percentile bootstrap
            # set abs=2.9 because there are lots of tests
            abs=2.9 / num_enrollments,
        ), l


def test_bayesian_bootstrap_vs_beta(spark_context_or_none):
    # The two distributions should be mathematically identical for binary data
    # like this; differences could emerge from
    # 1. implementation errors
    # 2. not taking enough bootstrap replications to suppress variance
    num_enrollments = 10000
    fake_data = pd.Series(np.zeros(num_enrollments))
    fake_data[:300] = 1

    boot_res = mabsbb.bootstrap_one_branch(fake_data, sc=spark_context_or_none)
    beta_res = mabsbin.summarize_one_branch_from_agg(
        pd.Series(
            {
                # `-1` to simulate Beta(0, 0) improper prior, closer to
                # bootstrap for quantiles (i think?)
                "num_enrollments": len(fake_data) - 1,
                "num_conversions": fake_data.sum() - 1,
            }
        )
    )

    for l in boot_res.index:
        assert boot_res.loc[l] == pytest.approx(
            beta_res.loc[l],
            # abs=1.9 is usually good enough with a percentile bootstrap
            # set abs=2.9 because there are lots of tests
            abs=2.9 / num_enrollments,
        ), (l, boot_res, beta_res)


def test_bayesian_bootstrap_vs_bootstrap_geometric(spark_context_or_none):
    num_enrollments = 20000

    rs = np.random.RandomState(42)
    data = rs.geometric(p=0.1, size=num_enrollments)

    bb_res = mabsbb.bootstrap_one_branch(data, sc=spark_context_or_none)
    pboot_res = mafsb.bootstrap_one_branch(data, sc=spark_context_or_none)

    assert bb_res["mean"] == pytest.approx(10, rel=1e-2)
    assert bb_res["0.5"] == pytest.approx(10, rel=1e-2)

    for l in bb_res.index:
        assert bb_res.loc[l] == pytest.approx(pboot_res.loc[l], rel=5e-3), (
            l,
            bb_res,
            pboot_res,
        )


def test_bayesian_bootstrap_vs_bootstrap_geometric_quantiles(spark_context_or_none):
    num_enrollments = 20000

    rs = np.random.RandomState(42)
    data = rs.geometric(p=0.1, size=num_enrollments)

    quantiles = [0.3, 0.5, 0.9]

    def calc_quantiles(x):
        return dict(zip(quantiles, np.quantile(x, quantiles)))

    bb_res = mabsbb.bootstrap_one_branch(
        data,
        stat_fn=mabsbb.make_bb_quantile_closure(quantiles),
        sc=spark_context_or_none,
    )
    pboot_res = mafsb.bootstrap_one_branch(
        data, stat_fn=calc_quantiles, sc=spark_context_or_none
    )

    for q in bb_res.index:
        for l in bb_res.columns:
            assert bb_res.loc[q, l] == pytest.approx(pboot_res.loc[q, l], rel=5e-3), (
                q,
                l,
                bb_res,
                pboot_res,
            )


def test_bayesian_bootstrap_vs_bootstrap_poisson(spark_context_or_none):
    num_enrollments = 10001

    rs = np.random.RandomState(42)
    data = rs.poisson(lam=10, size=num_enrollments)

    bb_res = mabsbb.bootstrap_one_branch(data, sc=spark_context_or_none)
    pboot_res = mafsb.bootstrap_one_branch(data, sc=spark_context_or_none)

    assert bb_res["mean"] == pytest.approx(10, rel=1e-2)
    assert bb_res["0.5"] == pytest.approx(10, rel=1e-2)

    for l in bb_res.index:
        assert bb_res.loc[l] == pytest.approx(pboot_res.loc[l], rel=5e-3), (
            l,
            bb_res,
            pboot_res,
        )


def test_bayesian_bootstrap_vs_bootstrap_poisson_quantiles(spark_context_or_none):
    num_enrollments = 10001

    rs = np.random.RandomState(42)
    data = rs.poisson(lam=10, size=num_enrollments)

    quantiles = [0.1, 0.5, 0.95]

    def calc_quantiles(x):
        return dict(zip(quantiles, np.quantile(x, quantiles)))

    bb_res = mabsbb.bootstrap_one_branch(
        data,
        stat_fn=mabsbb.make_bb_quantile_closure(quantiles),
        sc=spark_context_or_none,
    )
    pboot_res = mafsb.bootstrap_one_branch(
        data, stat_fn=calc_quantiles, sc=spark_context_or_none
    )

    for q in bb_res.index:
        for l in bb_res.columns:
            assert bb_res.loc[q, l] == pytest.approx(pboot_res.loc[q, l], rel=5e-3), (
                q,
                l,
                bb_res,
                pboot_res,
            )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np
import pytest

from mozanalysis.utils import add_days, all_, any_, date_sub, filter_outliers


def test_all_(spark):
    df = spark.createDataFrame(
        data=[
            [True, False, False],
            [True, False, True],
            [True, False, False],
            [True, False, True],
        ],
        schema=["all_true", "all_false", "mixed"],
    )
    pres = df.select(
        all_([df.all_true, df.all_true, df.all_true]).alias("true_1"),
        all_([df.all_true, df.all_false]).alias("false_2"),
        all_([df.all_false, df.all_false]).alias("false_3"),
        all_([df.mixed, df.all_false]).alias("false_4"),
        all_([df.mixed, df.all_true]).alias("mixed_5"),
    ).toPandas()
    assert pres.shape == (4, 5)
    assert not pres.isnull().any().any()
    assert pres["true_1"].all()
    assert not pres["false_2"].any()
    assert not pres["false_3"].any()
    assert not pres["false_4"].any()
    assert not pres["mixed_5"][::2].any()
    assert pres["mixed_5"][1::2].all()

    # Check this workaround is still necessary:
    with pytest.raises(ValueError):
        all([df.all_true, df.all_true])


def test_any_(spark):
    df = spark.createDataFrame(
        data=[
            [True, False, False],
            [True, False, True],
            [True, False, False],
            [True, False, True],
        ],
        schema=["all_true", "all_false", "mixed"],
    )
    pres = df.select(
        any_([df.all_true, df.all_true, df.all_true]).alias("true_1"),
        any_([df.all_true, df.all_false]).alias("true_2"),
        any_([df.all_false, df.all_false]).alias("false_3"),
        any_([df.mixed, df.all_true]).alias("true_4"),
        any_([df.mixed, df.all_false]).alias("mixed_5"),
    ).toPandas()
    assert pres.shape == (4, 5)
    assert not pres.isnull().any().any()
    assert pres["true_1"].all()
    assert pres["true_2"].any()
    assert not pres["false_3"].any()
    assert pres["true_4"].any()
    assert not pres["mixed_5"][::2].any()
    assert pres["mixed_5"][1::2].all()

    # Check this workaround is still necessary:
    with pytest.raises(ValueError):
        any([df.all_true, df.all_true])


def test_add_days():
    assert add_days("2019-01-01", 0) == "2019-01-01"
    assert add_days("2019-01-01", 1) == "2019-01-02"
    assert add_days("2019-01-01", -1) == "2018-12-31"


def test_date_sub():
    assert date_sub("2019-01-01", "2019-01-01") == 0
    assert date_sub("2019-01-02", "2019-01-01") == 1
    assert date_sub("2019-01-01", "2019-01-02") == -1


def test_filter_outliers():
    data = np.arange(100) + 1

    filtered = filter_outliers(data, 0.99)
    assert len(filtered) == 99
    assert filtered.max() == 99
    assert data.max() == 100


def test_filter_outliers_2():
    data = np.ones(100)

    filtered = filter_outliers(data, 0.99)
    assert len(filtered) == 100



import re


def rstify_links(s):
    return re.sub(r"\[(.*?)\]\((.*?)\)", r"`\1 <\2>`_", s)


# """unit tests"""
assert rstify_links("[foo](https://foo.bar)") == "`foo <https://foo.bar>`_"
assert (
    rstify_links("[foo bar](https://example.com?frob) baz")
    == "`foo bar <https://example.com?frob>`_ baz"
)


def process(app, what, name, obj, options, lines):
    if not hasattr(obj, "description"):
        return
    for i, line in enumerate(lines):
        if "..." in line:
            lines[i] = line.replace("...", rstify_links(obj.description))


def setup(app):
    app.connect("autodoc-process-docstring", process)
    return {
        "version": "1.0.0",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np
import pandas as pd
import pytest

import mozanalysis.bayesian_stats as mabs


def test_summarize_one_branch_samples():
    s = pd.Series(np.linspace(0, 1, 1001))

    res = mabs.summarize_one_branch_samples(s, [0.05, 0.31, 0.95])
    assert res.shape == (4,)
    assert res["0.05"] == pytest.approx(0.05)
    assert res["0.31"] == pytest.approx(0.31)
    assert res["0.95"] == pytest.approx(0.95)
    assert res["mean"] == pytest.approx(0.5)


def test_summarize_one_branch_samples_batch():
    s = pd.Series(np.linspace(0, 1, 1001))
    df = pd.DataFrame({"a": s, "b": s + 1})
    res = mabs.summarize_one_branch_samples(df, quantiles=[0.05, 0.31, 0.95])
    assert res.shape == (2, 4)

    assert res.loc["a", "0.05"] == pytest.approx(0.05)
    assert res.loc["a", "0.31"] == pytest.approx(0.31)
    assert res.loc["a", "0.95"] == pytest.approx(0.95)
    assert res.loc["a", "mean"] == pytest.approx(0.5)

    assert res.loc["b", "0.05"] == pytest.approx(1.05)
    assert res.loc["b", "0.31"] == pytest.approx(1.31)
    assert res.loc["b", "0.95"] == pytest.approx(1.95)
    assert res.loc["b", "mean"] == pytest.approx(1.5)


def test_summarize_joint_samples_trivial():
    quantiles = (0.05, 0.31, 0.95)
    res = mabs.summarize_joint_samples(
        pd.Series([6, 6, 6]), pd.Series([3, 3, 3]), quantiles=quantiles
    )
    assert res[("rel_uplift", "exp")] == 1.0
    assert res[("abs_uplift", "exp")] == 3.0
    assert res[("prob_win", None)] == 1
    assert res[("max_abs_diff", "0.95")] == 3.0

    assert res[("rel_uplift", "0.05")] == 1.0
    assert res[("rel_uplift", "0.31")] == 1.0
    assert res[("rel_uplift", "0.95")] == 1.0

    assert res[("abs_uplift", "0.05")] == 3.0
    assert res[("abs_uplift", "0.31")] == 3.0
    assert res[("abs_uplift", "0.95")] == 3.0


def test_summarize_joint_samples_batch_trivial():
    quantiles = (0.05, 0.31, 0.95)
    res = mabs.summarize_joint_samples(
        pd.DataFrame({"a": [6, 6, 6], "b": [1, 1, 1]}, columns=["a", "b"]),
        pd.DataFrame({"a": [3, 3, 3], "b": [1, 1, 1]}, columns=["b", "a"]),
        quantiles=quantiles,
    )
    assert res.loc["a", ("rel_uplift", "exp")] == 1.0
    assert res.loc["a", ("abs_uplift", "exp")] == 3.0
    assert res.loc["a", ("prob_win", None)] == 1
    assert res.loc["a", ("max_abs_diff", "0.95")] == 3.0

    assert res.loc["a", ("rel_uplift", "0.05")] == 1.0
    assert res.loc["a", ("rel_uplift", "0.31")] == 1.0
    assert res.loc["a", ("rel_uplift", "0.95")] == 1.0

    assert res.loc["a", ("abs_uplift", "0.05")] == 3.0
    assert res.loc["a", ("abs_uplift", "0.31")] == 3.0
    assert res.loc["a", ("abs_uplift", "0.95")] == 3.0

    assert res.loc["b", ("rel_uplift", "exp")] == 0.0
    assert res.loc["b", ("abs_uplift", "exp")] == 0.0
    assert res.loc["b", ("max_abs_diff", "0.95")] == 0.0

    assert res.loc["b", ("rel_uplift", "0.05")] == 0.0
    assert res.loc["b", ("rel_uplift", "0.31")] == 0.0
    assert res.loc["b", ("rel_uplift", "0.95")] == 0.0

    assert res.loc["b", ("abs_uplift", "0.05")] == 0.0
    assert res.loc["b", ("abs_uplift", "0.31")] == 0.0
    assert res.loc["b", ("abs_uplift", "0.95")] == 0.0



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from mozanalysis.metrics import DataSource, Metric

baseline = DataSource(
    name="baseline",
    from_expr="""(
                SELECT
                    p.*,
                    DATE(p.submission_timestamp) AS submission_date
                FROM `moz-fx-data-shared-prod.{dataset}.baseline` p
            )""",
    client_id_column="client_info.client_id",
    experiments_column_type="glean",
    default_dataset="org_mozilla_ios_focus",
)


events = DataSource(
    name="events",
    from_expr="""(
                SELECT
                    p.* EXCEPT (events),
                    DATE(p.submission_timestamp) AS submission_date,
                    event
                FROM
                    `moz-fx-data-shared-prod.{dataset}.events` p
                CROSS JOIN
                    UNNEST(p.events) AS event
            )""",
    client_id_column="client_info.client_id",
    experiments_column_type="glean",
    default_dataset="org_mozilla_ios_focus",
)

metrics = DataSource(
    name="metrics",
    from_expr="""(
                SELECT
                    p.*,
                    DATE(p.submission_timestamp) AS submission_date
                FROM `moz-fx-data-shared-prod.{dataset}.metrics` p
            )""",
    client_id_column="client_info.client_id",
    experiments_column_type="glean",
    default_dataset="org_mozilla_ios_focus",
)

baseline_ping_count = Metric(
    name="baseline_ping_count",
    data_source=baseline,
    select_expr="COUNT(document_id)",
    friendly_name="Baseline pings",
    description="Counts the number of `baseline` pings received from each client.",
)

metric_ping_count = Metric(
    name="metric_ping_count",
    data_source=metrics,
    select_expr="COUNT(document_id)",
    friendly_name="Metrics pings",
    description="Counts the number of `metrics` pings received from each client.",
)

first_run_date = Metric(
    name="first_run_date",
    data_source=baseline,
    select_expr="MIN(client_info.first_run_date)",
    friendly_name="First run date",
    description="The earliest first-run date reported by each client.",
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from textwrap import dedent

from mozanalysis.metrics import agg_any
from mozanalysis.segments import Segment, SegmentDataSource

#: SegmentDataSource: The clients_last_seen table.
clients_last_seen = SegmentDataSource(
    name="clients_last_seen",
    from_expr="mozdata.telemetry.clients_last_seen",
    window_start=0,
    window_end=0,
)

#: Segment: ...
regular_users_v3 = Segment(
    name="regular_users_v3",
    data_source=clients_last_seen,
    select_expr=agg_any("is_regular_user_v3"),
    friendly_name="Regular users (v3)",
    description=dedent(
        """\
        Clients who used Firefox on at least 14 of the 27 days prior to enrolling.
        This segment is characterized by high retention.
        """
    ),
)

#: Segment: ...
new_or_resurrected_v3 = Segment(
    name="new_or_resurrected_v3",
    data_source=clients_last_seen,
    select_expr="LOGICAL_OR(COALESCE(is_new_or_resurrected_v3, TRUE))",
    friendly_name="New or resurrected users (v3)",
    description=dedent(
        """\
        Clients who used Firefox on none of the 27 days prior to enrolling.
        """
    ),
)

#: Segment: ...
weekday_regular_v1 = Segment(
    name="weekday_regular_v1",
    data_source=clients_last_seen,
    select_expr=agg_any("is_weekday_regular_v1"),
    friendly_name="Weekday regular users (v1)",
    description=dedent(
        """\
        A subset of "regular users" who typically use Firefox on weekdays.
        """
    ),
)

#: Segment: ...
allweek_regular_v1 = Segment(
    name="allweek_regular_v1",
    data_source=clients_last_seen,
    select_expr=agg_any("is_allweek_regular_v1"),
    friendly_name="All-week regulars (v1)",
    description=dedent(
        """\
        A subset of "regular users" that have used Firefox on weekends.
        """
    ),
)

#: Segment: ...
new_unique_profiles = Segment(
    name="new_unique_profiles",
    data_source=clients_last_seen,
    select_expr="COALESCE(ANY_VALUE(first_seen_date) >= submission_date, TRUE)",
    friendly_name="New unique profiles",
    description=dedent(
        """\
        Clients that enrolled the first date their client_id ever appeared
        in telemetry (i.e. new, unique profiles).
    """
    ),
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import pandas as pd
import pytest
import scipy.stats as st

import mozanalysis.bayesian_stats.survival_func as mabssf


def test_get_thresholds():
    # One nonzero data point per percentile: 1 ... 101
    # Reverse order to make life slightly more interesting
    data = pd.Series(range(102)[::-1])
    res = mabssf.get_thresholds(data)

    assert len(data) - 1 == 101
    assert len(res) == 101

    for d, r in zip(data.sort_values()[:-1], res):
        assert r == pytest.approx(d)


def test_get_thresholds_2():
    data = pd.Series(index=range(1000)).fillna(0)
    data[30:40] = 1
    data[874] = 2
    data[300] = 4
    data[0] = 9001

    res = mabssf.get_thresholds(data)

    assert set(res) == {0, 1, 2, 4}


def test_one_thresh():
    df = pd.DataFrame(columns=["branch", "val"], index=range(1000))
    df.iloc[::2, 0] = "control"
    df.iloc[1::2, 0] = "test"
    df.fillna(0, inplace=True)
    df.iloc[:300, 1] = range(300)

    # Odd threshold: both branches have equal amounts of data above/below
    # Specifically, 129 above, 371 below
    res = mabssf._one_thresh(41, df, "val", "control")

    assert res["individual"]["test"].loc["0.025"] == pytest.approx(
        st.beta(129 + 1, 371 + 1).ppf(0.025), abs=1e-6
    )
    assert res["individual"]["test"].loc["0.5"] == pytest.approx(
        st.beta(129 + 1, 371 + 1).ppf(0.5), abs=1e-6
    )
    assert (
        res["individual"]["control"].loc["0.025"]
        == res["individual"]["test"].loc["0.025"]
    )
    assert (
        res["individual"]["control"].loc["0.5"] == res["individual"]["test"].loc["0.5"]
    )

    assert res["individual"]["test"].loc["0.5"] == pytest.approx(
        1.0 * (df.val > 41).sum() / len(df.val), abs=0.001
    )

    assert "_tmp_threshold_val" not in df.columns

    # Even threshold: 'control' has one fewer data point above it
    res2 = mabssf._one_thresh(42, df, "val", "control")
    assert res2["individual"]["test"].loc["0.025"] == pytest.approx(
        st.beta(129 + 1, 371 + 1).ppf(0.025), abs=1e-6
    )
    assert res2["individual"]["test"].loc["0.5"] == pytest.approx(
        st.beta(129 + 1, 371 + 1).ppf(0.5), abs=1e-6
    )
    assert (
        res2["individual"]["control"].loc["0.025"]
        < res2["individual"]["test"].loc["0.025"]
    )
    assert (
        res2["individual"]["control"].loc["0.5"] < res2["individual"]["test"].loc["0.5"]
    )

    assert "_tmp_threshold_val" not in df.columns


def test_compare_branches():
    df = pd.DataFrame(columns=["branch", "val"], index=range(1000), dtype="float")
    df.iloc[::2, 0] = "control"
    df.iloc[1::2, 0] = "test"
    df.iloc[:300, 1] = range(300)

    with pytest.raises(ValueError):
        mabssf.compare_branches(df, "val", thresholds=[0.0, 15.9])

    df.iloc[300:, 1] = 0
    res = mabssf.compare_branches(df, "val", thresholds=[0.0, 15.9])

    assert res["individual"]["control"].loc[0.0, "0.025"] == pytest.approx(
        st.beta(149 + 1, 351 + 1).ppf(0.025), abs=1e-6
    )
    assert res["individual"]["control"].loc[0.0, "0.5"] == pytest.approx(
        0.298537, abs=1e-6
    )

    assert res["individual"]["test"].loc[15.9, "0.025"] == pytest.approx(
        st.beta(142 + 1, 358 + 1).ppf(0.025), abs=1e-6
    )
    assert res["individual"]["test"].loc[15.9, "0.5"] == pytest.approx(
        0.284575, abs=1e-6
    )


def test_few_auto_thresholds():
    df = pd.DataFrame(columns=["branch", "val"], index=range(1000))
    df.iloc[::2, 0] = "control"
    df.iloc[1::2, 0] = "test"
    df.fillna(0, inplace=True)
    df.iloc[20:30] = 1
    df.iloc[405] = 100

    res = mabssf.compare_branches(df, "val")
    assert set(res["individual"]["test"].index) == {0, 1}
    assert set(res["individual"]["control"].index) == {0, 1}



import pytest
from cheap_lint import sql_lint

import mozanalysis.metrics.desktop as mmd
import mozanalysis.metrics.fenix as mmf
import mozanalysis.metrics.firefox_ios as mmios
import mozanalysis.metrics.focus_android as mmfoc
import mozanalysis.metrics.focus_ios as iosfoc
import mozanalysis.metrics.klar_android as mmk
import mozanalysis.metrics.klar_ios as iosk
from mozanalysis.metrics import DataSource, Metric

from . import enumerate_included


def test_imported_ok():
    assert mmd.active_hours
    assert mmf.uri_count
    assert mmios.baseline_ping_count
    assert mmfoc.metric_ping_count
    assert iosfoc.metric_ping_count
    assert mmk.baseline_ping_count
    assert iosk.baseline_ping_count


@pytest.fixture()
def included_metrics():
    return enumerate_included((mmd, mmf, mmios, mmk, mmfoc), Metric)


@pytest.fixture()
def included_datasources():
    return enumerate_included((mmd, mmf, mmios, mmk, mmfoc), DataSource)


def test_sql_not_detectably_malformed(included_metrics, included_datasources):
    for _, m in included_metrics:
        sql_lint(m.select_expr.format(experiment_slug="slug"))

    for _, ds in included_datasources:
        sql_lint(ds.from_expr_for(None))


def test_consistency_of_metric_and_variable_names(included_metrics):
    for name, metric in included_metrics:
        assert name == metric.name, metric


def test_included_metrics_have_docs(included_metrics):
    for _, m in included_metrics:
        assert m.friendly_name, m.name
        assert m.description, m.name



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np
import pandas as pd
import pytest

import mozanalysis.bayesian_stats.bayesian_bootstrap as mabsbb


def test_resample_and_agg_once():
    assert mabsbb._resample_and_agg_once(
        np.array([3.0]), np.array([3]), mabsbb.bb_mean
    ) == pytest.approx(3.0)


def test_resample_and_agg_once_multistat(stack_depth=0):
    data_values = np.array([0, 1])
    data_counts = np.array([10000, 10000])
    res = mabsbb._resample_and_agg_once(
        data_values,
        data_counts,
        lambda x, y: {
            "min": np.min(x),
            "max": np.max(x),
            "mean": np.dot(x, y),
        },
    )

    assert res["min"] == 0
    assert res["max"] == 1
    assert res["mean"] == pytest.approx(0.5, rel=1e-1)

    if stack_depth >= 3:
        assert res["mean"] != 0.5  # Extremely unlikely
    elif res["mean"] == 0.5:
        # This is a 0.5% event - implausible but not impossible.
        # Re-roll the dice a few times to make sure this was a fluke.
        test_resample_and_agg_once_multistat(stack_depth + 1)


def test_resample_and_agg_once_bcast(spark_context):
    b_data_values = spark_context.broadcast(np.array([3.0]))
    b_data_counts = spark_context.broadcast(np.array([3]))
    assert (
        mabsbb._resample_and_agg_once_bcast(
            b_data_values, b_data_counts, mabsbb.bb_mean, 42
        )
        == 3.0
    )


def test_get_bootstrap_samples(spark_context):
    res = mabsbb.get_bootstrap_samples(
        np.array([3.0, 3.0, 3.0]), num_samples=2, sc=spark_context
    )
    assert res.shape == (2,)
    assert res[0] == pytest.approx(3.0)
    assert res[1] == pytest.approx(3.0)


def test_get_bootstrap_samples_no_spark():
    test_get_bootstrap_samples(None)


def test_get_bootstrap_samples_multistat(spark_context, stack_depth=0):
    data = np.concatenate([np.zeros(10000), np.ones(10000)])
    res = mabsbb.get_bootstrap_samples(
        data,
        lambda x, y: {
            "min": np.min(x),
            "max": np.max(x),
            "mean": np.dot(x, y),
        },
        num_samples=2,
        sc=spark_context,
    )

    assert res.shape == (2, 3)

    assert (res["min"] == 0).all()
    assert (res["max"] == 1).all()
    assert res["mean"].iloc[0] == pytest.approx(np.mean(data), rel=1e-1)
    assert res["mean"].iloc[1] == pytest.approx(np.mean(data), rel=1e-1)

    # If we stuff up (duplicate) the seeds then things aren't random
    assert res["mean"].iloc[0] != res["mean"].iloc[1]

    if stack_depth >= 3:
        assert (res["mean"] != np.mean(data)).any()  # Extremely unlikely
    elif (res["mean"] == np.mean(data)).any():
        # Re-roll the dice a few times to make sure this was a fluke.
        test_get_bootstrap_samples_multistat(spark_context, stack_depth + 1)


def test_get_bootstrap_samples_multistat_no_spark():
    test_get_bootstrap_samples_multistat(None)


def test_bootstrap_one_branch(spark_context):
    data = np.concatenate([np.zeros(10000), np.ones(10000)])
    res = mabsbb.bootstrap_one_branch(
        data, num_samples=100, summary_quantiles=(0.5, 0.61), sc=spark_context
    )

    assert res["mean"] == pytest.approx(0.5, rel=1e-1)
    assert res["0.5"] == pytest.approx(0.5, rel=1e-1)
    assert res["0.61"] == pytest.approx(0.5, rel=1e-1)


def test_bootstrap_one_branch_no_spark():
    test_bootstrap_one_branch(None)


def test_bootstrap_one_branch_multistat(spark_context):
    data = np.concatenate([np.zeros(10000), np.ones(10000), [1e20]])
    res = mabsbb.bootstrap_one_branch(
        data,
        stat_fn=lambda x, y: {
            "max": np.max(x),
            "mean": np.dot(x, y),
        },
        num_samples=5,
        summary_quantiles=(0.5, 0.61),
        threshold_quantile=0.9999,
        sc=spark_context,
    )

    assert res.shape == (2, 3)

    assert res.loc["max", "mean"] == 1
    assert res.loc["max", "0.5"] == 1
    assert res.loc["max", "0.61"] == 1
    assert res.loc["mean", "mean"] == pytest.approx(0.5, rel=1e-1)
    assert res.loc["mean", "0.5"] == pytest.approx(0.5, rel=1e-1)
    assert res.loc["mean", "0.61"] == pytest.approx(0.5, rel=1e-1)


def test_bootstrap_one_branch_multistat_no_spark():
    test_bootstrap_one_branch_multistat(None)


def test_compare_branches(spark_context_or_none):
    data = pd.DataFrame(index=range(60000), columns=["branch", "val"], dtype="float")
    data.iloc[::3, 0] = "control"
    data.iloc[1::3, 0] = "same"
    data.iloc[2::3, 0] = "bigger"

    data.iloc[::2, 1] = 0
    data.iloc[1::2, 1] = 1

    data.iloc[2::12, 1] = 1

    assert data.val[data.branch != "bigger"].mean() == 0.5
    assert data.val[data.branch == "bigger"].mean() == pytest.approx(0.75)

    res = mabsbb.compare_branches(data, "val", num_samples=2, sc=spark_context_or_none)

    assert res["individual"]["control"]["mean"] == pytest.approx(0.5, rel=1e-1)
    assert res["individual"]["same"]["mean"] == pytest.approx(0.5, rel=1e-1)
    assert res["individual"]["bigger"]["mean"] == pytest.approx(0.75, rel=1e-1)

    assert "control" not in res["comparative"].keys()
    assert res["comparative"]["same"][("rel_uplift", "exp")] == pytest.approx(
        0, abs=0.1
    )
    assert res["comparative"]["bigger"][("rel_uplift", "exp")] == pytest.approx(
        0.5, abs=0.1
    )

    # num_samples=2 so only 3 possible outcomes
    assert res["comparative"]["same"][("prob_win", None)] in (0, 0.5, 1)
    assert res["comparative"]["bigger"][("prob_win", None)] == pytest.approx(
        1, abs=0.01
    )


def test_compare_branches_multistat(spark_context_or_none):
    data = pd.DataFrame(index=range(60000), columns=["branch", "val"], dtype="float")
    data.iloc[::3, 0] = "control"
    data.iloc[1::3, 0] = "same"
    data.iloc[2::3, 0] = "bigger"

    data.iloc[::2, 1] = 0
    data.iloc[1::2, 1] = 1

    data.iloc[2::12, 1] = 1

    assert data.val[data.branch != "bigger"].mean() == 0.5
    assert data.val[data.branch == "bigger"].mean() == pytest.approx(0.75)

    res = mabsbb.compare_branches(
        data,
        "val",
        stat_fn=lambda x, y: {
            "max": np.max(x),
            "mean": np.dot(x, y),
        },
        num_samples=2,
        sc=spark_context_or_none,
    )

    assert res["individual"]["control"].loc["mean", "mean"] == pytest.approx(
        0.5, rel=1e-1
    )
    assert res["individual"]["same"].loc["mean", "mean"] == pytest.approx(0.5, rel=1e-1)
    assert res["individual"]["bigger"].loc["mean", "mean"] == pytest.approx(
        0.75, rel=1e-1
    )

    assert "control" not in res["comparative"].keys()

    assert res["comparative"]["same"].loc[
        "mean", ("rel_uplift", "exp")
    ] == pytest.approx(0, abs=0.1)
    assert res["comparative"]["bigger"].loc[
        "mean", ("rel_uplift", "exp")
    ] == pytest.approx(0.5, abs=0.1)

    # num_samples=2 so only 3 possible outcomes
    assert res["comparative"]["same"].loc["mean", ("prob_win", None)] in (0, 0.5, 1)
    assert res["comparative"]["bigger"].loc[
        "mean", ("prob_win", None)
    ] == pytest.approx(1, abs=0.01)

    assert res["comparative"]["same"].loc["max", ("rel_uplift", "exp")] == 0
    assert res["comparative"]["bigger"].loc["max", ("rel_uplift", "exp")] == 0


def test_bb_mean():
    values = np.array([0, 1, 2])
    weights = np.array([0.1, 0.4, 0.5])

    assert mabsbb.bb_mean(values, weights) == pytest.approx(1.4)


def test_bb_quantile():
    values = np.array([0, 1, 2])
    weights = np.array([0.1, 0.4, 0.5])

    calc_median = mabsbb.make_bb_quantile_closure(0.5)

    assert calc_median(values, weights) == pytest.approx(1)

    calc_a_bunch = mabsbb.make_bb_quantile_closure([0, 0.1, 0.5, 1])

    res = calc_a_bunch(values, weights)
    assert res[0] == pytest.approx(0)
    assert res[0.1] == pytest.approx(0)
    assert res[0.5] == pytest.approx(1)
    assert res[1] == pytest.approx(2)



import pytest

from mozanalysis.metrics import DataSource


@pytest.mark.parametrize("experiments_column_type", [None, "simple", "native"])
def test_datasource_constructor_succeeds(experiments_column_type):
    DataSource(
        name="foo",
        from_expr="my_table.name",
        experiments_column_type=None,
    )


@pytest.mark.parametrize(
    "name,from_expr,experiments_column_type,error",
    [
        (None, "mytable", "simple", TypeError),
        ("name", None, "simple", TypeError),
        ("name", "mytable", "wrong", ValueError),
    ],
)
def test_datasource_constructor_fails(name, from_expr, experiments_column_type, error):
    with pytest.raises(error):
        DataSource(
            name=name,
            from_expr=from_expr,
            experiments_column_type=experiments_column_type,
        )


def test_complains_about_template_without_default():
    with pytest.raises(ValueError):
        DataSource(
            name="foo",
            from_expr="moz-fx-data-shared-prod.{dataset}.foo",
        )
    DataSource(
        name="foo",
        from_expr="moz-fx-data-shared-prod.{dataset}.foo",
        default_dataset="dataset",
    )



#!/usr/bin/env python
# encoding: utf-8

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from setuptools import setup, find_packages

tests_require = [
    "mock",
    "pytest",
    "pytest-black",
    "pytest-cov",
    "pytest-timeout",
    "pyspark",
]

docs_require = ["Sphinx", "sphinx-autobuild", "sphinx-rtd-theme"]

setup(
    name="mozanalysis",
    use_scm_version=True,
    author="Mozilla Corporation",
    author_email="fx-data-dev@mozilla.org",
    description="A library for Mozilla experiments analysis",
    url="https://github.com/mozilla/mozanalysis",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=[
        "attrs",
        "numpy",
        "pandas",
        "pyarrow",
        "scipy",
        "google-cloud-bigquery",
        "google-cloud-bigquery-storage",
    ],
    setup_requires=["pytest-runner", "setuptools_scm"],
    extras_require={
        "testing": tests_require,
        "docs": docs_require,
    },
    tests_require=tests_require,
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional

import attr


@attr.s(frozen=True, slots=True)
class SegmentDataSource:
    """Represents a table or view, from which segments may be defined.

    ``window_start`` and ``window_end`` define the window of data used
    to determine whether each client fits a segment. Ideally this
    window ends at/before the moment of enrollment, so that user's
    branches can't bias the segment assignment. ``window_start`` and
    ``window_end`` are integers, representing the number
    of days before or after enrollment.


    Args:
        name (str): Name for the Data Source. Should be unique to avoid
            confusion.
        from_expr (str): FROM expression - often just a fully-qualified
            table name. Sometimes a subquery. May contain the string
            ``{dataset}`` which will be replaced with an app-specific
            dataset for Glean apps. If the expression is templated
            on dataset, default_dataset is mandatory.
        window_start (int, optional): See above.
        window_end (int, optional): See above.
        client_id_column (str, optional): Name of the column that
            contains the ``client_id`` (join key). Defaults to
            'client_id'.
        submission_date_column (str, optional): Name of the column
            that contains the submission date (as a date, not
            timestamp). Defaults to 'submission_date'.
        default_dataset (str, optional): The value to use for
            `{dataset}` in from_expr if a value is not provided
            at runtime. Mandatory if from_expr contains a
            `{dataset}` parameter.
    """

    name = attr.ib(validator=attr.validators.instance_of(str))
    _from_expr = attr.ib(validator=attr.validators.instance_of(str))
    window_start = attr.ib(default=0, type=int)
    window_end = attr.ib(default=0, type=int)
    client_id_column = attr.ib(default="client_id", type=str)
    submission_date_column = attr.ib(default="submission_date", type=str)
    default_dataset = attr.ib(default=None, type=Optional[str])

    @default_dataset.validator
    def _check_default_dataset_provided_if_needed(self, attribute, value):
        self.from_expr_for(None)

    def from_expr_for(self, dataset: Optional[str]) -> str:
        """Expands the ``from_expr`` template for the given dataset.
        If ``from_expr`` is not a template, returns ``from_expr``.

        Args:
            dataset (str or None): Dataset name to substitute
                into the from expression.
        """
        effective_dataset = dataset or self.default_dataset
        if effective_dataset is None:
            try:
                return self._from_expr.format()
            except Exception as e:
                raise ValueError(
                    f"{self.name}: from_expr contains a dataset template but no value was provided."  # noqa:E501
                ) from e
        return self._from_expr.format(dataset=effective_dataset)

    def build_query(
        self,
        segment_list,
        time_limits,
        experiment_slug,
        from_expr_dataset=None,
    ):
        """Return a nearly self contained SQL query.

        The query takes a list of ``client_id``s from
        ``raw_enrollments``, and adds one non-NULL boolean column per
        segment: True if the client is in the segment, False otherwise.
        """
        return """SELECT
            e.client_id,
            e.branch,
            {segments}
        FROM raw_enrollments e
            LEFT JOIN {from_expr} ds
                ON ds.{client_id} = e.client_id
                AND ds.{submission_date} BETWEEN
                    DATE_ADD('{first_enrollment}', interval {window_start} day)
                    AND DATE_ADD('{last_enrollment}', interval {window_end} day)
                AND ds.{submission_date} BETWEEN
                    DATE_ADD(e.enrollment_date, interval {window_start} day)
                    AND DATE_ADD(e.enrollment_date, interval {window_end} day)
        GROUP BY e.client_id, e.branch""".format(
            client_id=self.client_id_column,
            submission_date=self.submission_date_column,
            from_expr=self.from_expr_for(from_expr_dataset),
            first_enrollment=time_limits.first_enrollment_date,
            last_enrollment=time_limits.last_enrollment_date,
            window_start=self.window_start,
            window_end=self.window_end,
            segments=",\n            ".join(
                f"{m.select_expr} AS {m.name}" for m in segment_list
            ),
        )

    @window_start.validator
    def window_start_lte_window_end(self, attribute, value):
        if value > self.window_end:
            raise ValueError("window_start must be <= window_end")


@attr.s(frozen=True, slots=True)
class Segment:
    """Represents an experiment Segment.

    Args:
        name (str): The segment's name; will be a column name.
        data_source (SegmentDataSource): Data source that provides
            the columns referenced in ``select_expr``.
        select_expr (str): A SQL select expression that includes
            an aggregation function (we ``GROUP BY client_id``).
            Returns a non-NULL ``BOOL``: ``True`` if the user is in the
            segment, ``False`` otherwise.
        friendly_name (str): A human-readable dashboard title for this segment
        description (str): A paragraph of Markdown-formatted text describing
            the segment in more detail, to be shown on dashboards
    """

    name = attr.ib(type=str)
    data_source = attr.ib(validator=attr.validators.instance_of(SegmentDataSource))
    select_expr = attr.ib(type=str)
    friendly_name = attr.ib(type=Optional[str], default=None)
    description = attr.ib(type=Optional[str], default=None)



def sql_lint(sql):
    safewords = [
        # Exceptions to skip linting
    ]
    for w in safewords:
        if w in sql:
            return

    # Check whether a python string template wasn't filled
    assert "{" not in sql
    assert "}" not in sql

    # Check for balanced parentheses
    assert sql.count("(") == sql.count(")")

    # Check for balanced quote marks
    assert sql.count("'") % 2 == 0



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# This allows wildcard imports: from mozanalysis import *
# Generally, import * should be avoided, however it is used by Jetstream
# We use it to expose modules available in mozanalysis package.
# https://github.com/mozilla/jetstream
import os

__all__ = []
basedir = os.path.dirname(__file__)
for root, subdirs, files in os.walk(basedir):
    __all__ += [module for module in subdirs if not module.startswith("_")]
    module_files = [
        file.replace(".py", "")
        for file in files
        if file.endswith(".py") and not file.startswith("_")
    ]
    if os.path.basename(root) in __all__:
        module_files = [os.path.basename(root) + "." + f for f in module_files]
    __all__ += module_files



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np
import pandas as pd

import mozanalysis.bayesian_stats as mabs
from mozanalysis.utils import filter_outliers


def compare_branches(
    df,
    col_label,
    ref_branch_label="control",
    stat_fn=np.mean,
    num_samples=10000,
    threshold_quantile=None,
    individual_summary_quantiles=mabs.DEFAULT_QUANTILES,
    comparative_summary_quantiles=mabs.DEFAULT_QUANTILES,
    sc=None,
):
    """Jointly sample bootstrapped statistics then compare them.

    Performs a percentile bootstrap, which, according to Efron,
    is not significantly more distasteful than a basic bootstrap,
    regardless of what you may read on Stack Overflow.

    Args:
        df: a pandas DataFrame of queried experiment data in the
            standard format (see ``mozanalysis.experiment``).
        col_label (str): Label for the df column contaning the metric
            to be analyzed.
        ref_branch_label (str, optional): String in ``df['branch']`` that
            identifies the branch with respect to which we want to
            calculate uplifts - usually the control branch.
        stat_fn (func, optional): A function that either:

            - Aggregates each resampled population to a scalar (e.g.
                the default, ``np.mean``), or
            - Aggregates each resampled population to a dict of
                scalars.

            In both cases, this function must accept a one-dimensional
            ndarray or pandas Series as its input.
        num_samples (int, optional): The number of bootstrap iterations
            to perform.
        threshold_quantile (float, optional): An optional threshold
            quantile, above which to discard outliers. E.g. `0.9999`.
        individual_summary_quantiles (list, optional): Quantiles to
            determine the confidence bands on individual branch
            statistics. Change these when making Bonferroni corrections.
        comparative_summary_quantiles (list, optional): Quantiles to
            determine the confidence bands on comparative branch
            statistics (i.e. the change relative to the reference
            branch, probably the control). Change these when making
            Bonferroni corrections.
        sc (optional): The Spark context, if available

    Returns a dictionary:
        If ``stat_fn`` returns a scalar (this is the default), then
        this function returns a dictionary has the following keys and
        values:

            'individual': dictionary mapping each branch name to a pandas
                Series that holds the expected value for the bootstrapped
                ``stat_fn``, and confidence intervals.
            'comparative': dictionary mapping each branch name to a pandas
                Series of summary statistics for the possible uplifts of
                the bootstrapped ``stat_fn`` relative to the reference branch.

        Otherwise, when ``stat_fn`` returns a dict, then this function
        returns a similar dictionary, except the Series are replaced with
        DataFrames. Each row in each DataFrame corresponds to one output
        of `stat_fn`, and is the Series that would be returned if ``stat_fn``
        computed only this statistic.
    """
    branch_list = df.branch.unique()

    if ref_branch_label not in branch_list:
        raise ValueError(
            "Branch label '{b}' not in branch list '{bl}".format(
                b=ref_branch_label, bl=branch_list
            )
        )

    samples = {
        # TODO: do we need to control seed_start? If so then we must be careful here
        b: get_bootstrap_samples(
            df[col_label][df.branch == b],
            stat_fn,
            num_samples,
            threshold_quantile=threshold_quantile,
            sc=sc,
        )
        for b in branch_list
    }

    return mabs.compare_samples(
        samples,
        ref_branch_label,
        individual_summary_quantiles,
        comparative_summary_quantiles,
    )


def bootstrap_one_branch(
    data,
    stat_fn=np.mean,
    num_samples=10000,
    seed_start=None,
    threshold_quantile=None,
    summary_quantiles=mabs.DEFAULT_QUANTILES,
    sc=None,
):
    """Run a bootstrap for one branch on its own.

    Resamples the data ``num_samples`` times, computes ``stat_fn`` for
    each sample, then returns summary statistics for the distribution
    of the outputs of ``stat_fn``.

    Args:
        data: The data as a list, 1D numpy array, or pandas Series
        stat_fn: Either a function that aggregates each resampled
            population to a scalar (e.g. the default value ``np.mean``
            lets you bootstrap means), or a function that aggregates
            each resampled population to a dict of scalars. In both
            cases, this function must accept a one-dimensional ndarray
            as its input.
        num_samples: The number of bootstrap iterations to perform
        seed_start: An int with which to seed numpy's RNG. It must
            be unique within this set of calculations.
        threshold_quantile (float, optional): An optional threshold
            quantile, above which to discard outliers. E.g. ``0.9999``.
        summary_quantiles (list, optional): Quantiles to determine the
            confidence bands on the branch statistics. Change these
            when making Bonferroni corrections.
        sc (optional): The Spark context, if available
    """
    samples = get_bootstrap_samples(
        data, stat_fn, num_samples, seed_start, threshold_quantile, sc
    )

    return mabs.summarize_one_branch_samples(samples, summary_quantiles)


def get_bootstrap_samples(
    data,
    stat_fn=np.mean,
    num_samples=10000,
    seed_start=None,
    threshold_quantile=None,
    sc=None,
):
    """Return ``stat_fn`` evaluated on resampled and original data.

    Do the resampling in parallel over the cluster.

    Args:
        data: The data as a list, 1D numpy array, or pandas series
        stat_fn: Either a function that aggregates each resampled
            population to a scalar (e.g. the default value ``np.mean``
            lets you bootstrap means), or a function that aggregates
            each resampled population to a dict of scalars. In both
            cases, this function must accept a one-dimensional ndarray
            as its input.
        num_samples: The number of samples to return
        seed_start: A seed for the random number generator; this
            function will use seeds in the range::

                [seed_start, seed_start + num_samples)

            and these particular seeds must not be used elsewhere
            in this calculation. By default, use a random seed.
        threshold_quantile (float, optional): An optional threshold
            quantile, above which to discard outliers. E.g. ``0.9999``.
        sc (optional): The Spark context, if available

    Returns:
        ``stat_fn`` evaluated over ``num_samples`` samples.

            * By default, a pandas Series of sampled means
            * if ``stat_fn`` returns a scalar, a pandas Series
            * if ``stat_fn`` returns a dict, a pandas DataFrame
              with columns set to the dict keys.
    """
    if not type(data) == np.ndarray:
        data = np.array(data.to_numpy(dtype='float', na_value=np.nan))

    if np.isnan(data).any():
        raise ValueError("'data' contains null values")

    if threshold_quantile:
        data = filter_outliers(data, threshold_quantile)

    if seed_start is None:
        seed_start = np.random.randint(np.iinfo(np.uint32).max)

    # Deterministic "randomness" requires careful state handling :(
    # Need to ensure every call has a unique, deterministic seed.
    seed_range = range(seed_start, seed_start + num_samples)

    if sc is None:
        summary_stat_samples = [
            _resample_and_agg_once(data, stat_fn, unique_seed)
            for unique_seed in seed_range
        ]

    else:
        try:
            broadcast_data = sc.broadcast(data)

            summary_stat_samples = (
                sc.parallelize(seed_range)
                .map(
                    lambda seed: _resample_and_agg_once_bcast(
                        broadcast_data=broadcast_data,
                        stat_fn=stat_fn,
                        unique_seed=seed % np.iinfo(np.uint32).max,
                    )
                )
                .collect()
            )

        finally:
            broadcast_data.unpersist()

    summary_df = pd.DataFrame(summary_stat_samples)
    if len(summary_df.columns) == 1:
        # Return a Series if stat_fn returns a scalar
        return summary_df.iloc[:, 0]

    # Else return a DataFrame if stat_fn returns a dict
    return summary_df


def _resample_and_agg_once_bcast(broadcast_data, stat_fn, unique_seed):
    return _resample_and_agg_once(broadcast_data.value, stat_fn, unique_seed)


def _resample_and_agg_once(data, stat_fn, unique_seed=None):
    random_state = np.random.RandomState(unique_seed)

    n = len(data)
    # TODO: can't we just use random_state.choice? Wouldn't that be faster?
    # There's not thaaat much difference in RAM requirements?
    randints = random_state.randint(0, n, n)
    resampled_data = data[randints]

    return stat_fn(resampled_data)
